package xyz.lychee.lagfixer.libs.adventure.key;

import java.util.Comparator;
import java.util.OptionalInt;
import java.util.stream.Stream;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.examination.Examinable;
import xyz.lychee.lagfixer.libs.examination.ExaminableProperty;

public interface Key extends Comparable<Key>, Examinable, Namespaced, Keyed {
   String MINECRAFT_NAMESPACE = "minecraft";
   char DEFAULT_SEPARATOR = ':';

   static @NotNull Key key(@KeyPattern final @NotNull String string) {
      return key(string, ':');
   }

   static @NotNull Key key(final @NotNull String string, final char character) {
      int index = string.indexOf(character);
      String namespace = index >= 1 ? string.substring(0, index) : "minecraft";
      String value = index >= 0 ? string.substring(index + 1) : string;
      return key(namespace, value);
   }

   static @NotNull Key key(final @NotNull Namespaced namespaced, @KeyPattern.Value final @NotNull String value) {
      return key(namespaced.namespace(), value);
   }

   static @NotNull Key key(@KeyPattern.Namespace final @NotNull String namespace, @KeyPattern.Value final @NotNull String value) {
      return new KeyImpl(namespace, value);
   }

   static @NotNull Comparator<? super Key> comparator() {
      return KeyImpl.COMPARATOR;
   }

   static boolean parseable(final @Nullable String string) {
      if (string == null) {
         return false;
      } else {
         int index = string.indexOf(58);
         String namespace = index >= 1 ? string.substring(0, index) : "minecraft";
         String value = index >= 0 ? string.substring(index + 1) : string;
         return parseableNamespace(namespace) && parseableValue(value);
      }
   }

   static boolean parseableNamespace(final @NotNull String namespace) {
      return !checkNamespace(namespace).isPresent();
   }

   static @NotNull OptionalInt checkNamespace(final @NotNull String namespace) {
      int i = 0;

      for(int length = namespace.length(); i < length; ++i) {
         if (!allowedInNamespace(namespace.charAt(i))) {
            return OptionalInt.of(i);
         }
      }

      return OptionalInt.empty();
   }

   static boolean parseableValue(final @NotNull String value) {
      return !checkValue(value).isPresent();
   }

   static @NotNull OptionalInt checkValue(final @NotNull String value) {
      int i = 0;

      for(int length = value.length(); i < length; ++i) {
         if (!allowedInValue(value.charAt(i))) {
            return OptionalInt.of(i);
         }
      }

      return OptionalInt.empty();
   }

   static boolean allowedInNamespace(final char character) {
      return KeyImpl.allowedInNamespace(character);
   }

   static boolean allowedInValue(final char character) {
      return KeyImpl.allowedInValue(character);
   }

   @KeyPattern.Namespace
   @NotNull String namespace();

   @KeyPattern.Value
   @NotNull String value();

   @NotNull String asString();

   default @NotNull String asMinimalString() {
      return this.namespace().equals("minecraft") ? this.value() : this.asString();
   }

   default @NotNull Stream<? extends ExaminableProperty> examinableProperties() {
      return Stream.of(ExaminableProperty.of("namespace", this.namespace()), ExaminableProperty.of("value", this.value()));
   }

   default int compareTo(final @NotNull Key that) {
      return comparator().compare(this, that);
   }

   default @NotNull Key key() {
      return this;
   }
}
